/*================================================================

File: client.c

Title: Client Module

Author: Ashok

Purpose:
This module communicates with the Java client

================================================================*/

#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>

#include "xpbasic.h"
#include "vrec2.h"
#include "player2.h"
#include "client.h"

#define NO_MESG "NONE"
#define NO_SIG 0
#define KILL_MESG "DIE"
#define KILL_SIG 1

#define FAKE_ADDR "229.229.229.229"
#define FAKE_PORT 22999

#define LOCKFILE "/tmp/lock.file"
#define SYSHDRFILE "/home/ashok/research/dsap/mpeg-c2/system_header.dat"

#define SERVER_SOCKET 9877

char SYSHDR[200];
int SYSHDR_LEN;

int prev_client_present(void);
void  send_to_prev_client(char *filename);
void new_client(char *filename);
unsigned long process_file(char *filename);
void sigterm_handler(int i);

int clientfd;

void client_setup(char *filename)
{
	clientfd=0;

	if (prev_client_present())
		send_to_prev_client(filename);
	else
		new_client(filename);
}

int prev_client_present(void)
{
	struct stat m;
	return stat(LOCKFILE, &m)!=-1;
}

void send_to_prev_client(char *filename)
{	
	int clientfd;
	struct sockaddr_in addr;
	int addrlen;
	unsigned long mcaddr;
	char buf[4];

	log("Reading multicast address from file");
	mcaddr = process_file(filename);

	log("Trying to send message to earlier client");	
	 
	clientfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (clientfd < 0)
		perr("Unable to create socket");

	memset(&addr, 0, sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=htonl(INADDR_ANY); /* N.B.: differs from sender */
	addr.sin_port=htons(SERVER_SOCKET);
	addrlen=sizeof(addr);

	buf[0] = (mcaddr >> 24);
	buf[1] = (mcaddr >> 16) & 0xff;
	buf[2] = (mcaddr >> 8) & 0xff;
	buf[3] = (mcaddr) & 0xff;

	if (sendto(clientfd, buf, 4, 0, (struct sockaddr *) &addr, addrlen) < 0)
	{
		perror("Unable to communicate with running client");
		return ;
	}

	log("Message sent to running client. Terminating");
	sleep(1);
	exit(1);
}

void new_client(char *filename)
{
	int fd;
	struct sockaddr_in addr; int addrlen;

	log("Installing new client");

	if ((clientfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		perr("Unable to create socket");

	/*
	if (fcntl(clientfd, F_SETFL, O_NONBLOCK) < 0)
		perr("Unable to make socket nonblocking");
		*/
	memset(&addr, 0, sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=htonl(INADDR_ANY); 
	addr.sin_port=htons(SERVER_SOCKET);
	addrlen=sizeof(addr);

  /* Bind the socket to the address */
	if (bind(clientfd,(struct sockaddr *) &addr,sizeof(addr)) < 0) 
		perr("Unable to BIND socket to address.");

	log("Client listener installed");

	fd = open(SYSHDRFILE, O_RDONLY);
	if (fd<0)
		perr("Unable to open headers file");

	if ((SYSHDR_LEN=read(fd, SYSHDR, 200))<0)
		perr("Unable to read headers from file");

	log("Read headers of %d bytes", SYSHDR_LEN);
	close(fd);

	init_video_socket();
	vrec_join(process_file(filename));

	/* The parent exits */
	if (fork()!=0)
	{
		log("Parent exiting");
		exit(1);
	}

	fd = open(LOCKFILE, O_CREAT | O_TRUNC | O_WRONLY, 0666);
	close(fd);

	atexit(client_exit);

	signal(SIGTERM, sigterm_handler);

}

unsigned long process_file(char *filename)
{
	static FILE *fd;
	char grp[64];
	
	log("Reading file %s", filename);
	fd = fopen(filename, "r");
	if (fd==NULL)
	{
		warn("Unable to open file %s", filename);
		return;
	}

	if (fscanf(fd, " %s", grp) <= 0)
	{
		warn("Unable to read multicast address");
		return;
	}

	if (!strcmp(grp, KILL_MESG))
		return KILL_SIG;
		
	if (!strcmp(grp, NO_MESG))
		return NO_SIG;
		
	return inet_addr(grp);
}

void client_do(void)
{
	unsigned long dat;
	unsigned char buf[4];
	if (recvfrom(clientfd, buf, 4, 0, 0, 0) < 0)
	{
		if (errno == EWOULDBLOCK)
		{
			return;
		}
		else
			perr("Error while reading");
	}

	dat = ((int)buf[0] << 24) + ((int)buf[1] << 16) + ((int)buf[2] << 8) + buf[3];

	if (dat==KILL_SIG)
	{
		log("Running client received an EXIT signal. Terminating\n");
		exit(0);
	}
	
	if (dat==NO_SIG)
	{
	    vrec_leave();
	    return;
	}

	msg("Joining address %d", dat);

	vrec_join(dat);
}

void client_exit(void)
{
	unlink(LOCKFILE);
}

void sigterm_handler(int i)
{
	exit(2);
}

