#include "netdefs.h"
#include "error.h"        /* err_xxx functions prototypes */
#include "tellwait.h"     /* TELL_xxx, WAIT_xxx functions prototypes */
#include "vcripc.h"       /* needed for session.h */
#include "newsess.h"      /* all other functions prototypes */
#include "path.h"         /* path for LOGFILE is in here */

extern char *pname;   /* Program name. Initialized to NULL in error.c */
extern FILE *logfp;   /* Log file pointer. Initialized to stderr in error.c */
                      /* Output of all err_xxx functions goes there */

int fildes, fld, *pid_val;
int sem_for_pid, shm_for_pid;
pid_t serv_pid, dsap_pid;


void cleanup();
void start_daemon();
void dsap_pid_ipcs();

main(int argc, char *argv[]) {
  pid_t pid;
  int c, debug = 0;

  int i = 0, fl = 1;
  pname = argv[0];  

		
  /* Check for -d (debug) option */
  while((c = getopt(argc, argv, "d")) != EOF) {
    switch(c) {
		case 'd':                                 /* debug */
			debug = 1;
			break;
		case '?':
			err_quit("%s: usage: %s [ -d ]", pname, pname);
    }
  }

	dsap_pid_ipcs();
	
  /* If not debug, start as a daemon */
  if(debug == 0)  
    start_daemon();  

  /* Initialize server */
  init_serv();
        
  TELL_WAIT();


  if((pid = fork()) < 0)  
    err_sys("%s: ERROR: fork error", pname);  

  else if(pid > 0) {       /* parent */
  
		for( ; ; ) {

			check_shmem();
			read_in_vec();

			TELL_CHILD();  
			WAIT_CHILD();  
  
			write_to_nw();
			
			TELL_CHILD();  
			WAIT_CHILD();
		}
  }

  else {                 /* child */

		for( ; ; ) {
			
			write_to_nw();
			
			WAIT_PARENT();
			TELL_PARENT();  
      
			check_shmem();  
			read_in_vec();

			WAIT_PARENT();
			TELL_PARENT();  
		}  
  }
} 


/*
 * Starts daemon.
 */

void start_daemon() {
	pid_t pid;
	int signo;

	/* Open log file */
	if((logfp = fopen(LOGFILE, "a")) == NULL) {
		fprintf(stderr, "%s: cannot open log file %s\n", pname, LOGFILE);
		exit(1);
	}
      
	err_msg("%s: attempting to start daemon @ %s", pname, tstamp());
      




	/* Install signal handler */
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	for(signo=3; signo<33; signo++)
		signal(signo, cleanup);
      
	/* Start forking */
	if((pid = fork()) < 0)
		err_sys("%s: ERROR: cannot fork first child for daemon", pname);
	else if(pid > 0)
		exit(0);                             /* parent */
      
	/* First child process */
	setsid();                              /* become session leader */
      
	if((pid = fork()) < 0)
		err_sys("%s: ERROR: cannot fork second child for daemon", pname);
	else if(pid > 0)
		exit(0);                             /* first child */
      
	/* Second child process */
	err_msg("%s: daemon started successfully @ %s", pname, tstamp());
	serv_pid = getpid();
      
	/* Close file descriptors */
	close(0);
	close(1);
	close(2);
      
	chdir("/");
	umask(0);
}

void dsap_pid_ipcs()
{
	/* Read the value of pid of the DSAP process */
	if((shm_for_pid = shmget(SHMKEY0, sizeof(int), PERMS)) < 0)
	{
		err_sys("Could not get shared memory");
	}
		
	if((pid_val = (int *) shmat(shm_for_pid, (char *) 0, 0)) == (int *)-1)
	{
		err_sys("Could not attach shared memory");
	}
	
	if((sem_for_pid = sem_open(SEMKEY0)) < 0)
	{
		err_sys("Could not open semaphore");
	}

	sem_wait(sem_for_pid);     /* Lock shared memory for write */

	dsap_pid = *pid_val;
	err_msg("DSAP pid : %d", dsap_pid);
	*pid_val = 1;
	
	bzero(pid_val+1, NSERVERS); /* Clear the shared memory for future use */
	
	sem_signal(sem_for_pid);   /* Unlock shared memory */

	if(shmdt((void *) pid_val) < 0)
	{
		err_sys("Could not detach shared memory");
	}

	/*
	if(shmctl(shm_for_pid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		err_sys("Could not detach shared memory");
	
	sem_rm(sem_for_pid);
	*/
}

/*
 * Handles signals.
 */
void cleanup(int signo)
{
     pid_t pid;

     /* Parent exits first, then child */
     if((pid = getpid()) != serv_pid)
       sleep(1);
     err_msg("\n%s: pid %d: exiting on signal %d @ %s", pname, pid,
	     signo, tstamp());
     exit(0); 
}




