#include "netdefs.h"
#include "vcripc.h"
#include "server.h"

/****************************************************************************/
/* Provide a simpler and easy to understand interface to the System V       */
/* semaphore system calls. There are 7 routines available to the user.      */
/*                                                                          */
/* id = sem_create(key, initval);  # create with initial value or open      */
/* id = sem_open(key);             # open (must already exist)              */
/* sem_wait(id);                   # wait = P = down by 1                   */
/* sem_signal(id);                 # signal = V = up by 1                   */
/* sem_op(id, amount);             # wait if (amount < 0)                   */
/*                                 # signal if (amount > 0)                 */
/* sem_close(id);                  # close                                  */ 
/* sem_rm(id);                     # remove (delete)                        */
/*                                                                          */ 
/* We create and use a 3-member set for the requested semaphore.            */ 
/* The first member, [0], is the actual semaphore value, and the second     */ 
/* member, [1], is a counter used to know when all processes have finished  */ 
/* with the semaphore. The counter is initialised to a large number,        */ 
/* decremented on every create or open and incremented on every close.      */ 
/* This way we can use the "adjust" feature provided by System V so that    */ 
/* any process that exits without calling sem_close() is accounted for.     */ 
/* It doesn't help us if the last process does this (as we have no way of   */
/* getting control to remove the semaphore) but it will work if any process */
/* other than the last does an exit (intentional or unintentional).         */
/*                                                                          */ 
/* The third member, [2], of the semaphore set is used as a lock variable   */ 
/* to avoid any race conditions in the sem_create() and sem_close()         */ 
/* functions.                                                               */ 
/*                                                                          */ 
/****************************************************************************/
#define BIGCOUNT 10000           /* Initial value of process counter */

/****************************************************************************/
/* Define the semaphore operation arrays for the semop() calls.             */
/****************************************************************************/
static struct sembuf op_lock[2] =
    {
     2, 0, 0,                  /* Wait for [2] (lock) to equal 0 */
     2, 1, SEM_UNDO            /* then increment [2] to 1 - this locks it */
                               /* UNDO to release the locks if processes  */
                               /* exit before explicitly unlocking        */
    };

static struct sembuf op_endcreate[2] =
    {
     1, -1, SEM_UNDO,          /* decrement [1] (proc counter) with undo  */
                               /* on exit. UNDO to adjust proc counter if */
                               /* process exits before explicitly calling */
                               /* sem_close()                             */
     2, -1, SEM_UNDO           /* then decrement [2] (lock) back to 0     */
    };

static struct sembuf op_open[1] =
    {
     1, -1, SEM_UNDO           /* decrement [1] (proc counter) with undo  */
                               /* on exit                                 */
    };

static struct sembuf op_close[3] =
    {
     2, 0, 0,                  /* wait for [2] (lock) to equal 0         */
     2, 1, SEM_UNDO,           /* then increment [2] to 1 - this locks it*/
     1, 1, SEM_UNDO            /* then increment [1] (proc counter)      */
    };

static struct sembuf op_unlock[1] =
    {
     2, -1, SEM_UNDO          /* decrement [2] (lock) back to 0         */
    };

static struct sembuf op_op[1] =
    {
     0, 99, SEM_UNDO         /* decrement or increment [0] with undo on */
                             /* exit. The 99 is set to the actual amount*/
                             /* to add or subtract ( +ve or -ve )       */
    };

#ifdef LINUX
union semun semctl_arg; 
#else
union {
    int val;
    struct semid_ds *buf;
    ushort *array;
} semctl_arg;
#endif


/****************************************************************************/
/* Create a semaphore with a specified initial value.                       */
/* If the semaphore already exists, we don't initialize it (of course).     */
/* We return the semaphore ID if all OK, else -1.                           */
/****************************************************************************/
int sem_create(key_t key, int initval)
    {
     register int id, semval;
	
     if(key == IPC_PRIVATE)
         return(-1);                /* Not intended for private semaphores */
     else if(key == (key_t) -1)
         return(-1);                /* Probably an ftok() error by caller  */

again:
     if((id = semget(key, 3, PERMS|IPC_CREAT)) < 0)
         return(-1);               /* Permission problem or tables full */

/****************************************************************************/
/* When the semaphore is created, we know that the value of all 3 members   */
/* is 0. Get a lock on the semaphore by waiting for [2] to equal 0, then    */
/* increment it.                                                            */
/*                                                                          */
/* There is a race condition here. There is a possibility that between the  */
/* semget() above and the semop() below, another process can call our       */
/* sem_close() function which can remove the semaphore if the process is the*/
/* last one using it. Therefore, we handle the error condition of an invalid*/
/* semaphore ID specially below, and if it does happen, we just go back and */
/* create it again.                                                         */
/****************************************************************************/
     if(semop(id, &op_lock[0], 2) < 0)
         {
          if(errno == EINVAL) goto again;
          logmessage("SEMOP", "ERR", "Can't lock");
         }

/****************************************************************************/
/* Get the value of the process counter. If it equals 0, then no one has    */
/* initialized the semaphore yet.                                           */
/****************************************************************************/
     if(semval = semctl(id, 1, GETVAL, 0) < 0)
          logmessage("SEMOP", "ERR", "Can't GETVAL");

     if(semval == 0)
         {
/****************************************************************************/
/* We could initialize by doing a SETALL, but that would clear the adjust   */
/* value that we set when we locked the semaphore above. Instaed, we'll do  */
/* system calls to initialize [0] and [1].                                  */
/****************************************************************************/
          semctl_arg.val = initval;
          if(semctl(id, 0, SETVAL, semctl_arg) < 0)
              logmessage("SEMOP", "ERR", "Can't SETVAL [0]");

          semctl_arg.val = BIGCOUNT;
          if(semctl(id, 1, SETVAL, semctl_arg) < 0)
              logmessage("SEMOP", "ERR", "Can't SETVAL [1]");
         }

/****************************************************************************/
/* Decrement the process counter and then release the lock.                 */
/****************************************************************************/
     if(semop(id, &op_endcreate[0], 2) < 0)
          logmessage("SEMOP", "ERR", "Can't end create");

     return(id);
    }/* End of sem_create() */

/***************************************************************************/
/* Open a semaphore that must already exist.                               */ 
/* This function should be used, instaed of sem_create() if the caller     */
/* knows that the semaphore must already be existing. Returns semaphore ID */
/* if all OK, else -1.                                                     */
/***************************************************************************/
int sem_open(key_t key)
    {
     register int id;

     if(key == IPC_PRIVATE)
         return(-1);            /* not intended for private semaphores */
     else if(key == (key_t) -1)
         return(-1);            /* probably an ftok() error by caller */

     if((id =semget(key, 3, 0)) < 0)
         return(-1);            /* Doesn't exist or tables full */

     if(semop(id, &op_open[0], 1) < 0) /* Decrement the process counter. */
                                       /* A lock is not needed for this. */
         logmessage("SEMOP", "ERR", "Can't open");

     return(id);
    }/* End of sem_open() */

/***************************************************************************/
/* Remove a semaphore.                                                     */ 
/* This call is intended to be called by the creator of the semaphore,so   */
/* that when it is being shut down an IPC_RMID is done on it, regardless   */
/* of what other processes may be using it. Most other processes would use */
/* sem_close() instead.                                                    */
/***************************************************************************/
int sem_rm(int id)
    {
     if(semctl(id, 0, IPC_RMID, 0) < 0)
         logmessage("SEMOP", "ERR", "Can't IPC_RMID");
    }/* End of sem_rm() */

/**************************************************************************/
/* Close a semaphore.                                                     */
/* Unlike the remove function above, this function is for a process to    */
/* call before it exits, when it is done with the semaphore.              */
/* We decrement the counter of processes using the semaphore, and if this */
/* was the last one, we acn remove the semaphore.                         */
/**************************************************************************/
int sem_close(int id)
    {
     register int semval;

     /* Get a lock on the semaphore and increment the process counter */
     if(semop(id, &op_close[0], 3) < 0)
         logmessage("SEMOP", "ERR", "Can't op_close");

/**************************************************************************/
/* Now that we have a lock, read the value of the process counter to see  */
/* if this is the last reference to the semaphore.                        */
/* There is a race condition here,as previously discussed in sem_create() */
/**************************************************************************/
     if((semval = semctl(id, 1, GETVAL, 0)) < 0)
         logmessage("SEMOP", "ERR", "Can't GETVAL");

     if(semval > BIGCOUNT)
         {
          logmessage("SEMOP", "ERR", "sem[1] > BIGCOUNT");
          exit(1);
         }
     else if(semval == BIGCOUNT)
         sem_rm(id);
     else if(semop(id, &op_unlock[0], 1) < 0)
         logmessage("SEMOP", "ERR", "Can't unlock");

    }/* End of sem_close() */

/**************************************************************************/
/* Wait until a semaphore's value is greater than 0, then decrement it by */
/* 1 and return.                                                          */
/* Djkistra's P operation. Tanenbaum's DOWN operation.                    */
/**************************************************************************/
int sem_wait(int id)
    {
     sem_op(id, -1);     /* Should be a macro, whaddaya think ? */
                         /* Will change later                   */
    }/* End of sem_wait() */

/**************************************************************************/
/* Increment a semaphore by 1.                                            */
/* Djkistra's V operation. Tanenbaum's UP operation.                      */
/**************************************************************************/
int sem_signal(int id)
    {
     sem_op(id, 1);      /* Should be a macro, whaddaya think ? */
                         /* Will change later                   */
    }/* End of sem_signal() */

/**************************************************************************/
/* General semaphore operation. Increment or decrement by a user-defined  */
/* non-zero amount.                                                       */ 
/**************************************************************************/
int sem_op(int id, int value)
    {
     if((op_op[0].sem_op = value) == 0)
         logmessage("SEMOP", "ERR", "Can't have value == 0");

     if(semop(id, &op_op[0], 1) < 0)
         logmessage("SEMOP", "ERR", "Error in op_op[0]");
    }/* End of sem_op() */
